@@ -43,12 +43,14 @@ class FE(object):
 
         self.cat_feat_combination = cat_feat_combination
 
-    def transform(self, df: pd.DataFrame, n_jobs=1):
-        """采样估计特征数
+    def transform(self, df: pd.DataFrame, n_jobs=1, threshold=10000):
+        """提前计算一下特征值数量，过滤
 
         @param df:
+        @param n_jobs:
         @return:
         """
+
         if self.cat_feats is None:
             self.cat_feats = self.infer_cat_feats(df)
             logger.info(f"Infer cat feats: {self.cat_feats}")
@@ -58,23 +60,49 @@ class FE(object):
                 logger.info(f"Infer num feats: {self.num_feats}")
 
         if self.cat_feat_combination is None:
-            self.cat_feat_combination = self.combination_all(self.cat_feats)
+            # self.cat_feat_combination = self.combination_all(self.cat_feats)
+            self.cat_feat_combination_df = self.calculate_featvalue_num(df, self.cat_feats, n_jobs, threshold)
+            self.cat_feat_combination = self.cat_feat_combination_df['keys'].tolist()
 
-        ##############################################################
-        # self.calculate_feature_number()
-        ##############################################################
+            ##############################################################
+            # self.calculate_feature_number()
+            ##############################################################
+
+        bar = tqdm(self.cat_feat_combination, desc='agg🐢')
 
-        bar = tqdm(self.cat_feat_combination, desc='🐢')
         if n_jobs == 1:
             for keys in bar:
                 df_ = self.group_calculate(df, keys)
-                df = df.merge(df_, on=keys)
+                df = df.merge(df_, 'left')  # df = df.merge(df_, on=keys)
+
         else:  # 多进程加速
             def func(keys):
-                return self.group_calculate(df, keys), keys
+                return self.group_calculate(df, keys)
 
-            for df_, keys in bar | xJobs(func, n_jobs):
-                df = df.merge(df_, on=keys)
+            dfs = bar | xJobs(func, n_jobs) | xtqdm(desc=f'joining🐢')
+
+            for df_ in dfs:
+                df = df.merge(df_, 'left')  # todo: 并行设计
+
+            #############################无加速效果#################################
+            # def func(dfs):
+            #     dff = df.copy()
+            #     for df_ in dfs:
+            #         dff = dff.merge(df_)
+            #     return dff
+            #
+            # def dfs_join(l, func=func, n_jobs=n_jobs):
+            #     batch_size = len(l) // n_jobs
+            #
+            #     if batch_size == 1:
+            #         return func(l)
+            #
+            #     l = l | xgroup(batch_size) | xtqdm(desc=f'joining🐢bs={batch_size}') | xJobs(func, n_jobs)
+            #
+            #     return dfs_join(l)
+            #
+            # df = dfs_join(dfs)
+            ##############################################################
 
         return df
 
@@ -82,14 +110,22 @@ class FE(object):
         agg_dict = self._get_agg_dict(keys)
         agg_dict[keys[0]] = ['count']  # count只统计一次
 
-        df_ = df.groupby(keys).agg(agg_dict)
+        df_ = df.groupby(keys, dropna=False).agg(agg_dict)
         df_.columns = [f"{':'.join(keys)}.{func.upper()}({feat})" for feat, func in df_.columns]
         df_.reset_index(inplace=True)
         return df_
 
     def _get_agg_dict(self, keys):
-        num_agg_dict = {feat: self.num_funcs for feat in self.num_feats}
-        cat_agg_dict = {feat: self.cat_funcs for feat in self.cat_feats if feat not in keys}
+        if self.num_funcs:
+            num_agg_dict = {feat: self.num_funcs for feat in self.num_feats}
+        else:
+            num_agg_dict = {}  # num_funcs=[] 不聚合num特征
+
+        if self.cat_funcs:
+            cat_agg_dict = {feat: self.cat_funcs for feat in self.cat_feats if feat not in keys}
+        else:
+            cat_agg_dict = {}  # cat_funcs=[] 不聚合cat特征
+
         return {**cat_agg_dict, **num_agg_dict, **self.agg_dict}
 
     @staticmethod
@@ -99,6 +135,30 @@ class FE(object):
             comb_list += map(list, combinations(s, i))
         return comb_list
 
+    def calculate_featvalue_num(self, df, cat_feats, n_jobs=6, threshold=10000):
+        func = lambda keys: (df[keys].drop_duplicates().__len__(), len(keys), keys)
+
+        _ = self.combination_all(cat_feats) | xtqdm(desc='计算特征值数🐢') | xJobs(func, n_jobs)
+        df_ = (
+            pd.DataFrame(_, columns=['num', 'keys_len', 'keys'])
+                .sort_values('num', ascending=False, ignore_index=True)[lambda df: df['num'] <= threshold]
+        )
+        return df_
+
+    # def calculate_featvalue_num(self, df, cat_feats, threshold=10000):
+    #
+    #     feat2num = df[cat_feats].nunique().to_dict()
+    #
+    #     df_ = pd.DataFrame(
+    #         [
+    #             (keys | xmap(lambda k: feat2num.get(k, 1)) | xreduce(lambda x, y: x * y), len(keys), keys)
+    #             for keys in self.combination_all(cat_feats)
+    #         ],
+    #         columns=['num', 'keys_len', 'keys']
+    #     ).sort_values('num', ascending=False, ignore_index=True)[lambda df: df['num'] <= threshold]
+    #
+    #     return df_
+
     @staticmethod
     def infer_cat_feats(df, threshold=64):
         threshold = min(len(df) // 100 + 8, threshold)
@@ -122,10 +182,11 @@ if __name__ == '__main__':
 
     df = pd.concat([df1, df2], 1)
 
-    # ft = FE(cat_feat_combination=[['N0'], ['N1'], ['N2']])
-    ft = FE()
+    # ft = FE()
+    # ft = FE(num_funcs=[])
+    ft = FE(cat_funcs=[], num_funcs=[])
 
-    df_ = ft.transform(df, n_jobs=4)
+    df_ = ft.transform(df, n_jobs=2)
 
     print(df_.shape)
 
